@@ -1,3 +1,4 @@
+use derive_more::{From, Into};
 use pyo3::{prelude::*, types::PyBytes};
 use serde::{Deserialize, Serialize};
 use solana_sdk::signer::{
@@ -35,7 +36,7 @@ mod keypair_serde {
 }
 
 #[pyclass(module = "solders.keypair", subclass)]
-#[derive(PartialEq, Debug, Serialize, Deserialize)]
+#[derive(PartialEq, Debug, Serialize, Deserialize, From, Into)]
 /// A vanilla Ed25519 key pair.
 ///
 /// Calling ``Keypair()`` creates a new, random ``Keypair``.
@@ -251,18 +252,6 @@ impl Default for Keypair {
     }
 }
 
-impl From<KeypairOriginal> for Keypair {
-    fn from(keypair: KeypairOriginal) -> Self {
-        Self(keypair)
-    }
-}
-
-impl From<Keypair> for KeypairOriginal {
-    fn from(k: Keypair) -> Self {
-        k.0
-    }
-}
-
 impl AsRef<KeypairOriginal> for Keypair {
     fn as_ref(&self) -> &KeypairOriginal {
         &self.0
